#include "RaveBase/RaveInterface/rave/BTagFactory.h"
#include "TrackingTools/TransientTrack/interface/TransientTrack.h"
#include "RecoVertex/VertexPrimitives/interface/TransientVertex.h"
#include "DataFormats/Candidate/interface/Particle.h"
#include "RaveBase/Converters/interface/CmsToRaveObjects.h"
#include "RaveBase/Converters/interface/RaveToCmsObjects.h"
#include "RaveBase/Converters/interface/RaveStreamers.h"
#include "RaveBase/Converters/interface/RaveComputations.h"
#include "RaveBase/RaveEngine/interface/RaveReconstructor.h"
#include "RaveBase/Converters/interface/MagneticFieldWrapper.h"
#include "RaveBase/RaveEngine/interface/RaveBTagger.h"
#include "RaveBase/RaveInterface/rave/RaveCalibration.h"
#include "FWCore/MessageLogger/interface/MessageLogger.h"
#include <iostream>
#include <iomanip>

using namespace rave;
using namespace std;

namespace {
  reco::Particle sum ( const vector < Track > & t )
  {
    double px=0.;
    double py=0.;
    double pz=0.;
    double E=0.;
    char Charge=0;
    for ( vector< Track >::const_iterator i=t.begin(); i!=t.end() ; ++i )
    {
      px+=i->state().momentum().x();
      py+=i->state().momentum().y();
      pz+=i->state().momentum().z();
      E+=0.; // all massless FIXME
    }
    return reco::Particle ( Charge, reco::Particle::LorentzVector ( px, py, pz, E ) );
  }
}

BTagFactory::BTagFactory ( const RaveReconstructor & r ) :
  theRector ( r.clone() ), theTagger ( 0 ) 
  
{
}

BTagFactory::~BTagFactory()
{
  if ( theRector ) delete theRector;
  if ( theTagger ) delete theTagger;
}

BTagFactory::BTagFactory ( const BTagFactory & o ) :
   theRector ( 0 ), theTagger ( 0 )
{
  // cout << "[rave::BTagFactory] copy constructor." << endl;
  if ( o.theRector ) theRector = o.theRector->clone();
  if ( o.theTagger ) theTagger = o.theTagger->clone();
}

double BTagFactory::tag ( const vector < Track > & trks, const Vertex & primary,
                      const Vector3D & jet ) const
{
  if ( !theTagger )
  {
    cout << "[BTagBTagFactory] error: no initialised tagger!!" << endl;
    return -1.0;
  }

  vector < reco::TransientTrack > ttrks;
  for ( vector< Track >::const_iterator i=trks.begin(); i!=trks.end() ; ++i )
  {
      ttrks.push_back ( reco::TransientTrack ( *i ) );
  }

  RaveToCmsObjects converter;
  TransientVertex vtx = converter.convert ( primary );
  reco::Particle p = sum ( trks );
  edm::LogInfo ("rave::BTagFactory") << " trying to tag with vtx at " << vtx.position();
  return theTagger->tag ( ttrks, vtx, p );
}

void BTagFactory::calibrate( const rave::Calibration & c ) const
{
  if ( !theTagger )
  {
    theTagger = new BTagger( c, theRector );
  } else {
    cout << "[rave::BTagFactory] have already got a btagger! Ignoring calibration!" << endl;
  }
}

BTagFactory::BTagFactory ( const VertexFactory & v ) :
  theRector ( v.reconstructor().clone() ), theTagger(0)
{}

rave::Calibration BTagFactory::getCalibration( ) const
{
  if ( theTagger )
  {
    return theTagger->getCalibration();
  }
  cout << "[BTagBTagFactory] no btagger available" << endl;
  exit(-1);
}

bool BTagFactory::hasBTagging() const
{
  return BTagger::hasBTagging();
}

void BTagFactory::teach ( const vector < Track > & trks, const Vertex & primary,
                      const Vector3D & jet, char tag )
{
  if ( !theTagger )
  {
    cout << "[BTagBTagFactory] error: no initialised tagger!!" << endl;
    return;
  }

  vector < reco::TransientTrack > ttrks;
  for ( vector< Track >::const_iterator i=trks.begin(); i!=trks.end() ; ++i )
  {
      ttrks.push_back ( reco::TransientTrack ( *i ) );
  }

  RaveToCmsObjects converter;
  TransientVertex vtx = converter.convert ( primary );
  reco::Particle p = sum ( trks );
  return theTagger->teach ( ttrks, vtx, p, tag );
}

