#include "RaveBase/RaveInterface/rave/VertexFactory.h"
#include "TrackingTools/TransientTrack/interface/TransientTrack.h"
#include "RecoVertex/VertexPrimitives/interface/TransientVertex.h"
#include "DataFormats/Candidate/interface/Particle.h"
#include "RaveBase/Converters/interface/CmsToRaveObjects.h"
#include "RaveBase/Converters/interface/RaveToCmsObjects.h"
#include "RaveBase/Converters/interface/RaveStreamers.h"
#include "RaveBase/Converters/interface/RaveComputations.h"
#include "RaveBase/RaveEngine/interface/RaveReconstructor.h"
#include "RaveBase/Converters/interface/MagneticFieldWrapper.h"
#include <iostream>
#include <iomanip>

using namespace rave;
using namespace std;

VertexFactory::VertexFactory ( const rave::MagneticField & field,
                   const rave::Propagator & prop,
                   const std::string & method,
                   int verbosity ) :
  theRector ( new RaveReconstructor ( method ) ), theMethod ( method ),
  theField ( field.copy() ) , theProp ( prop.copy() ),
  theBeamspot ( Ellipsoid3D() ), 
  theVerbosity ( verbosity )
  
{
    setup();
}

VertexFactory::VertexFactory ( const rave::MagneticField & field,
                   const rave::Propagator & prop,
                   const rave::Ellipsoid3D & beamspot,
                   const std::string & method,
                   int verbosity ) :
  theRector ( new RaveReconstructor ( method ) ), theMethod ( method ),
  theField ( field.copy() ) , theProp ( prop.copy() ),
  theBeamspot ( beamspot ), theVerbosity ( verbosity )
{
    setup();
}

void VertexFactory::setup()
{
    MagneticFieldWrapper::Instance()->registry ( *theField );
    if (!theRector)
    {
      cout << "[rave::VertexFactory] error: could not initialize reconstructor: " << theMethod
           << endl;
    }

    if ( theVerbosity > 20 )
    {
      cout << "[rave::VertexFactory] magnetic field at (0,0,0) is "
           << MagneticFieldWrapper::Instance()->inTesla ( GlobalPoint (0.,0.,0.) )
           << endl;
    }
}

VertexFactory::~VertexFactory()
{
    if ( theField ) delete theField;
    if ( theProp) delete theProp;
    if ( theRector ) delete theRector;
    for ( map < string, RaveReconstructor * >::const_iterator i=theRectors.begin(); 
          i!=theRectors.end() ; ++i )
    {
      delete i->second;
    }
    theRectors.clear();
}

vector < Vertex > VertexFactory::create ( const vector < Track > & trks, bool use_bs ) const
{
    if ( !theRector )
    {
      cout << "[VertexVertexFactory] error: trying to use uninitialised factory!"
           << endl;
      return vector < Vertex > ();
    }
    return fit ( trks, *theRector, use_bs );
}

vector < Vertex > VertexFactory::create ( const vector < Track > & trks, const string & method, bool use_bs ) const
{
  RaveReconstructor * tmp = theRectors[method];
  if ( !tmp )
  {
    tmp = new RaveReconstructor ( method );
    if (!tmp)
    {
      cout << "[VertexVertexFactory] error cannot initialse \"" << method << "\"" << endl;
      return vector < Vertex > ();
    }
    theRectors[method] = tmp;
  }
  return fit ( trks, *(tmp), use_bs );
}

VertexFactory VertexFactory::operator= ( const VertexFactory & o )
{
  // cout << "[rave::VertexFactory] assignment op! theProp=" << o.theProp << endl;
  if ( o.theProp && o.theField )
  {
    return VertexFactory ( *(o.theField), *(o.theProp), o.theMethod, o.theVerbosity );
  }
  return VertexFactory();
}

VertexFactory::VertexFactory ( const VertexFactory & o ) :
   theRector ( 0 ), theMethod ( o.theMethod ), theField ( 0 ),
   theProp ( 0 ), theVerbosity ( o.theVerbosity )
{
  // cout << "[rave::VertexFactory] copy constructor." << endl;
  if ( o.theRector ) theRector = o.theRector->clone();
  if ( o.theField ) theField = o.theField->copy();
  if ( o.theProp ) theProp = o.theProp->copy();
  for ( map < string, RaveReconstructor *  >::const_iterator i=o.theRectors.begin();
        i!=theRectors.end() ; ++i )
  {
    theRectors[i->first]=i->second->clone();
  }
}

vector < Vertex > VertexFactory::fit ( const vector < Track > & trks, const RaveReconstructor & r,
   bool use_bs ) const
{
    vector < rave::Vertex > ret;
    vector < reco::TransientTrack > ttrks;
    for ( vector< Track >::const_iterator i=trks.begin(); i!=trks.end() ; ++i )
    {
        ttrks.push_back ( reco::TransientTrack ( *i ) );
    }

    vector < TransientVertex > tmp = r.vertices ( ttrks );
    CmsToRaveObjects converter;
    for ( vector< TransientVertex >::const_iterator i=tmp.begin(); 
          i!=tmp.end() ; ++i )
    {
      ret.push_back ( converter.convert (*i, trks ) );
    }

    if ( theVerbosity>1 )
    {
        cout << "[rave::VertexFactory] produced " << ret.size() << " vertices from " << ttrks.size()
                  << " tracks." << endl;
    };
    return ret;
}

const int VertexFactory::verbosity() const
{
  return theVerbosity;
}

const std::string & VertexFactory::method() const
{
  return theMethod;
}

bool VertexFactory::hasBeamspot() const
{
  return theBeamspot.isValid();
}

const rave::Ellipsoid3D & VertexFactory::getBeamspot() const
{
  return theBeamspot;
}

const rave::Propagator & VertexFactory::getPropagator() const
{
    return *theProp;
}

const rave::MagneticField & VertexFactory::getMagneticField() const
{
    return *theField;
}

const RaveReconstructor & VertexFactory::reconstructor() const
{
  return (*theRector);
}
